// TestHD.cpp : Does a basic Amiga native harddisk test
//

#include "windows.h"
#include <stdlib.h>
#include <stdio.h>

#include "..\..\ADFlib\Include\adflib.h"
#pragma comment(lib, "..\\..\\ADFlib\\Libs\\adf.lib")


void Callback_Verbose(char *msg)
{
  printf("ADFLib: Verbose [%s]\n", msg);
}
void Callback_Error(char *msg)
{
  printf("ADFLib: Error [%s]\n", msg);
}
void Callback_Warning(char *msg)
{
  printf("ADFLib: Warning [%s]\n", msg);
}


void PrintWindowsError()
{
   DWORD dwError = ::GetLastError();
   LPVOID lpvMessageBuffer;
   DWORD nLen = ::FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
                   FORMAT_MESSAGE_IGNORE_INSERTS |
                   FORMAT_MESSAGE_FROM_SYSTEM,
                   NULL, 
                   dwError,
                   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                   (LPSTR)&lpvMessageBuffer, 
                   0, NULL);
   LPSTR pstrMessage = (LPSTR)lpvMessageBuffer;
   printf("WinNT: Error %ld, [%s]\n", dwError, pstrMessage);
   // Free the buffer allocated by the system.
   ::LocalFree(lpvMessageBuffer);
}

BOOL ReadSector(HANDLE hDrv, long iSect, int iSize, void *lpvoidBuf)
{
  void *lpvoidTempBuf;
  DWORD dwActual;

  lpvoidTempBuf = ::VirtualAlloc(NULL, 512, MEM_COMMIT, PAGE_READWRITE);

  if (::SetFilePointer(hDrv, iSect * 512, NULL, FILE_BEGIN) == 0xFFFFFFFF) {
    PrintWindowsError();
    printf("Error: Attempt to set filepointer failed.\n");
    ::VirtualFree(lpvoidTempBuf, 0, MEM_RELEASE);
    return FALSE;
  }

  if (!::ReadFile(hDrv, lpvoidTempBuf, 512, &dwActual, NULL)) {
    PrintWindowsError();
    printf("Error: Attempt to read 1st sector failed.\n");
    ::VirtualFree(lpvoidTempBuf, 0, MEM_RELEASE);
    return FALSE;
  }

  memcpy(lpvoidBuf, lpvoidTempBuf, iSize);
  ::VirtualFree(lpvoidTempBuf, 0, MEM_RELEASE);

  return TRUE;
}

BOOL DumpSector(BYTE *pBuf)
{
#define vchr(x) (x<32 ? ' ' : x)
  for( int i=0, x=0; i<4; i++, x+=8 ) {
  printf("  %02X %02X %02X %02X %02X %02X %02X %02X  %c%c%c%c%c%c%c%c\n", 
    pBuf[x+0],pBuf[x+1],pBuf[x+2],pBuf[x+3],pBuf[x+4],pBuf[x+5],pBuf[x+6],pBuf[x+7],
    vchr(pBuf[x+0]),vchr(pBuf[x+1]),vchr(pBuf[x+2]),vchr(pBuf[x+3]),vchr(pBuf[x+4]),vchr(pBuf[x+5]),vchr(pBuf[x+6]),vchr(pBuf[x+7]));
  }
#undef vchr
  return TRUE;
}

BOOL GetDriveGeometry(HANDLE hDevice)
{
  DISK_GEOMETRY pdg = { 0 };      // disk drive geometry structure
  BOOL bResult;                   // results flag
  DWORD junk;                     // discard results

  bResult = ::DeviceIoControl(hDevice,  // device we are querying
    IOCTL_DISK_GET_DRIVE_GEOMETRY,      // operation to perform
    NULL, 0,                            // no input buffer, so pass zero
    &pdg, sizeof(pdg),                  // output buffer
    &junk,                              // discard count of bytes returned
    (LPOVERLAPPED) NULL);               // synchronous I/O
  if (!bResult) {  
    // IOCTL failed
    PrintWindowsError();
    printf("Error: Attempt to get drive geometry failed.\n");
    return FALSE;
  }

  printf("  Cylinders = %I64d\n", pdg.Cylinders);
  printf("  Tracks per cylinder = %ld\n", (ULONG) pdg.TracksPerCylinder);
  printf("  Sectors per track = %ld\n", (ULONG) pdg.SectorsPerTrack);
  printf("  Bytes per sector = %ld\n", (ULONG) pdg.BytesPerSector);

  LONGLONG DiskSize;
  DiskSize = pdg.Cylinders.QuadPart * (ULONG)pdg.TracksPerCylinder * (ULONG)pdg.SectorsPerTrack * (ULONG)pdg.BytesPerSector;
  printf("  Disk size = %I64d (Bytes) = %I64d (Mb)\n", DiskSize, DiskSize / (1024 * 1024));

  return TRUE;
}

BOOL TestPhysicalDrive(LPSTR pstrDeviceName)
{
  printf("Opening physical device: %s\n", pstrDeviceName);

  HANDLE hDevice;                                 // handle to the drive to be examined 
  hDevice = ::CreateFile(pstrDeviceName,
             GENERIC_READ | GENERIC_WRITE,        // need both read/write access
             FILE_SHARE_READ | FILE_SHARE_WRITE,  // share mode
             NULL,                                // default security attributes
             OPEN_EXISTING,                       // disposition
             0,                                   // file attributes
             NULL);                               // don't copy any file's attributes
  if (hDevice == INVALID_HANDLE_VALUE) { 
    // we can't open the drive
    PrintWindowsError();
    printf("Error: Attempt to open device failed.\n");
    return FALSE;
  }

  printf("Getting drive geometry\n");
  BOOL bResult;
  bResult = GetDriveGeometry(hDevice);
  if( !bResult ) {
    ::CloseHandle(hDevice);
    return FALSE;
  }

  printf("Locking device\n");
  DWORD dwRet;	
  bResult = ::DeviceIoControl(hDevice, FSCTL_LOCK_VOLUME, NULL, 0, NULL, 0, &dwRet, NULL);
  if( !bResult ) {
    PrintWindowsError();
    printf("Error: Attempt to lock device failed.\n");
    ::CloseHandle(hDevice);
    return FALSE;
  }

  printf("Reading sector 0\n");
  BYTE buf[512];
  for(int i=0; i<512; i++) buf[i]=0xCC;
  bResult = ReadSector(hDevice, 0, 512, &buf);
  if( !bResult ) {
    ::DeviceIoControl(hDevice, FSCTL_UNLOCK_VOLUME, NULL, 0, NULL, 0, &dwRet, NULL);
    ::CloseHandle(hDevice);
    return FALSE;
  }
  DumpSector(buf);

  printf("Unlocking device\n");
  bResult = ::DeviceIoControl(hDevice, FSCTL_UNLOCK_VOLUME, NULL, 0, NULL, 0, &dwRet, NULL);
  if( !bResult ) {
    PrintWindowsError();
    printf("Error: Attempt to unlock device failed.\n");
    ::CloseHandle(hDevice);
    return FALSE;
  }

  printf("Closing handles\n");
  ::CloseHandle(hDevice);     // we're done with the handle

  return TRUE;
}

BOOL PrintDeviceInfo(Device *dev)
{
  printf("  Cylinders = %ld, Heads = %ld, Sectors = %ld\n", dev->cylinders, dev->heads,dev->sectors);
  printf("  Volumes = %d\n",dev->nVol);
  return TRUE;
}

BOOL PrintVolumeInfo(Volume *vol)
{
  if (vol->volName!=NULL) printf("  Volume Name: \"%s\"\n", vol->volName);
  printf("  Placed between sectors [%ld-%ld]\n",vol->firstBlock, vol->lastBlock);

  printf("  Flags: %s ",isFFS(vol->dosType) ? "FFS" : "OFS");
  if (isINTL(vol->dosType)) printf(" INTL");
  if (isDIRCACHE(vol->dosType)) printf(" DIRCACHE");
  printf("\n");

  printf("  Filled at %2.1f%%\n", 100.0-(adfCountFreeBlocks(vol)*100.0)/(vol->lastBlock - vol->firstBlock +1) );
  return TRUE;
}

BOOL PrintFileInformation(Entry* entry)
{
  if (entry->type==ST_LFILE || entry->type==ST_LDIR || entry->type==ST_LSOFT) return TRUE;
  if (entry->type==ST_DIR) printf("  %s/\n",entry->name); else printf("  %s\n",entry->name);
  return TRUE;
}

BOOL PrintDirectoryInformation(Volume *vol)
{
  List *list, *tree;
  tree = list = adfGetRDirEnt(vol,vol->curDirPtr,TRUE);
  while(tree) {
    Entry* entry;
    entry = (Entry*)tree->content;
    PrintFileInformation(entry);
    tree = tree->next;
  }
  adfFreeDirList(list);

  return TRUE;
}

BOOL TestADF(int nDrive, int nPartition, BOOL bUseCache)
{
  printf("Initializing ADFlib\n");
 
  adfEnvInitDefault(); // initialize the library
  adfChgEnvProp(PR_VFCT, Callback_Verbose);          
  adfChgEnvProp(PR_WFCT, Callback_Warning);          
  adfChgEnvProp(PR_EFCT, Callback_Error);          

  printf("ADFlib version: %s (%s)\n", adfGetVersionNumber(),adfGetVersionDate());

  CHAR szDevName[64];
  ::wsprintf(szDevName, "|H%d", nDrive);
  //::wsprintf(szDevName, "C:\\Temp\\Test.adf");

  printf("Mounting the device: %s\n", szDevName);
  Device *dev;
  dev = adfMountDev( szDevName, TRUE );
  if( dev==NULL ) {
    printf("Error: Unable to mount device\n");
    return FALSE;
  }

  if (nPartition >= dev->nVol) {
    printf("Error: This device has only %d partition(s)\n", dev->nVol);
    adfUnMountDev(dev);
    adfEnvCleanUp();
    return FALSE;
  }

  printf("Getting device information\n");
  PrintDeviceInfo(dev);

  printf("Mounting the volume: %d\n", nPartition);
  Volume *vol;
  vol = adfMount(dev, nPartition, TRUE);
  if (!vol) {
    printf("Error: Cannot mount the volume\n");
    adfUnMountDev(dev);
    adfEnvCleanUp();
    return FALSE;
  }

  printf("Getting volume information\n");
  PrintVolumeInfo(vol);

  if (bUseCache && isDIRCACHE(vol->dosType) ) {
    BOOL bFlag = TRUE;  
    adfChgEnvProp(PR_USEDIRC,&bFlag);
  }

  printf("Getting Root directory information\n");
  PrintDirectoryInformation(vol);

  printf("Closing system\n");
  adfUnMount(vol);
  adfUnMountDev(dev);
  adfEnvCleanUp();

  return TRUE;
}

void PrintUsage()
{
  printf("Usage:\n");
  printf("TestHD <drive> [-p <partition>][-c]\n");
  printf(" -p x   Use partition x (0 is first partition, 1 is second)\n");
  printf(" -c     Use caching if possible\n");
  printf("\n");
}

int main(int argc, char* argv[])
{
  printf("TestHD - Amiga HardDisk test utility\n\n");

  __try
  {

    if( argc==1 ) {
      PrintUsage();
      return 0;
    }

    int nDrive = atol(argv[1]);
    if( nDrive<0 ) {
      printf("Error: Invalid drive number\n");
      return 0;
    }

    printf("Running basic WindowsNT drive test... \n");
    CHAR szName[64];
    ::wsprintf(szName, "\\\\.\\PhysicalDrive%d", nDrive);
    TestPhysicalDrive(szName);
    printf("Done\n\n");

    printf("Parsing arguments...\n\n");
    int nPartition = 0;
    BOOL bUseCache = FALSE;
    for( int i=2; i<argc; i++ ) {
      if( argv[i][0]=='-' ) {
        switch( argv[i][1] ) {
        case 'p':
          nPartition = atol(argv[i+1]);
          break;
        case 'c':
          bUseCache = TRUE;
          break;
        }
      }
    }

    printf("Running ADFlib drive test...\n");
    TestADF(nDrive, nPartition, bUseCache);
    printf("Done\n\n");

  }
  __except( GetExceptionCode() == STATUS_ACCESS_VIOLATION )
  {
    printf("Error: GPF occured (%ld)! Reboot computer!", GetExceptionCode());
  }

  return 0;
}
